#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Korean Sentence Splitter
# Split Korean text into sentences using heuristic algorithm.
#
# Copyright (C) 2021 Hyun-Woong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import gc
import math
from typing import List, Union, Tuple
from .cache import LRUCache
from .morph import MorphExtractor

from .base import (
    Const,
    Eojeol,
    Postprocessor,
    Preprocessor,
    check_pos,
    do_push_pop_symbol,
    empty,
    length_constraints,
    get_chunk_with_index,
    preprocess_text,
    build_preprocessed_list,
)
from .rule import Table, Stats, ID

_morph = MorphExtractor()
# _cache = LRUCache(max_size=100)


class KSS:
    def __init__(
        self,
        use_heuristic: bool = False,
        use_quotes_brackets_processing: bool = False,
        max_recover_step: int = 5,
        max_recover_length: int = 20000,
        backend: str = "fugashi",
        dicdir=None,
        userdic_path=None,
        # disable_gc: bool = False,
        # cache_size: int = 1000,
        **kwargs,
    ) -> Union[List[str], List[List[str]]]:
        """
        Initialize KSS.

        Args:
                use_heuristic (bool): use heuristic algorithms or not
                use_quotes_brackets_processing (bool): use quotes or bracket processing or not
                max_recover_step (int): maximum step for quote and bracket misalignment recovering
                max_recover_length (int): maximum text length to recover when quote and bracket misaligned
                backend (str): max length of text to use morpheme feature
                disable_gc (bool): disable garbage collecting (It helps to improve speed)
        """
        assert isinstance(backend, str), "param `backend` must be `str` type"

        self.backend = backend.lower()

        assert self.backend in [
            "mecab-python3",
            "fugashi",
            "none",
        ], "Wrong backend! Currently, we support [`fugashi`, `mecab-python3`, `none`] backend."

        assert isinstance(
            use_heuristic, bool
        ), "param `use_heuristic` must be `bool` type"
        self.use_heuristic = use_heuristic
        assert isinstance(
            use_quotes_brackets_processing, bool
        ), "param `use_quotes_brackets_processing` must be `bool` type"
        self.use_quotes_brackets_processing = use_quotes_brackets_processing
        assert isinstance(
            max_recover_step, int
        ), "param `max_recover_step` must be `int` type"
        self.max_recover_step = max_recover_step
        assert isinstance(
            max_recover_length, int
        ), "param `max_recover_length` must be `int` type"
        self.max_recover_length = max_recover_length
        # assert isinstance(num_workers, int), "param `num_workers` must be `int` type"
        # self.num_workers = get_num_workers(num_workers)

        # self.disable_gc = disable_gc
        # if self.disable_gc:
        # 	gc.disable()
        _morph.parse_args(
            backend=self.backend, dicdir=dicdir, userdic_path=userdic_path
        )
        # if cache_size > 1:
        # 	_cache.max_size = cache_size

    def segment(
        self,
        text: str,
    ) -> List[str]:
        """
        Split document to sentences.

        Args:
                text (str): input text

        Returns:
                List[str]: list of segmented sentences
        """
        assert isinstance(text, str), "param `text` must be str."

        results = []

        self.max_recover_step = length_constraints(
            text,
            self.max_recover_length,
            self.max_recover_step,
        )

        preprocessed_list = build_preprocessed_list(text)
        if len(preprocessed_list) == 0:
            return results

        for t in preprocessed_list:
            results += self._split_into_sentences(
                text=t,
                use_heuristic=self.use_heuristic,
                use_quotes_brackets_processing=self.use_quotes_brackets_processing,
                max_recover_step=self.max_recover_step,
                max_recover_length=self.max_recover_length,
                backend=self.backend,
            )

        # if self.disable_gc:
        # 	gc.enable()

        return results

    def chunk(
        self,
        text: str,
        max_length: int,
        overlap: bool = False,
    ) -> List[str]:
        """
        Split chunks from input texts by max_length.

        Args:
                text (Union[str, List[str], tuple]): input texts
                max_length (int): max length of ecah chunk
                overlap (bool): whether allow duplicated sentence

        Returns:
                Union[List[str], List[List[str]]]: chunks of segmented sentences
        """

        assert isinstance(text, str), "param `text` must be str."
        assert isinstance(max_length, int), "param `max_length` must be `int` type."
        assert isinstance(overlap, bool), "param `overlap` must be `bool` type."

        span, chunks = [], []
        text = preprocess_text(text)

        for index in self._split_into_sentence_index(text):
            if len(span) > 0:
                if index[0] - span[0][1] > max_length:
                    chunks.append(get_chunk_with_index(text, span))
                    if overlap:
                        span = span[math.trunc(len(span) / 2) :]
                    else:
                        span = []

            span.append(index)
        chunks.append(get_chunk_with_index(text, span))
        return chunks

    def _split_into_sentence_index(self, text) -> List[Tuple[int, int]]:
        sentences = self.segment(text)
        offset, sentence_indice = 0, []

        for sentence in sentences:
            sentence_indice.append(
                (
                    offset + text.index(sentence),
                    offset + text.index(sentence) + len(sentence),
                )
            )
            offset += text.index(sentence) + len(sentence)
            text = text[text.index(sentence) + len(sentence) :]

        return sentence_indice

    def _split_into_sentences(
        self,
        text: str,
        use_heuristic: bool,
        use_quotes_brackets_processing: bool,
        max_recover_step: int,
        max_recover_length: int,
        backend: str,
        recover_step: int = 0,
    ):
        # _cache = self._cache
        # _morph = self._morph
        # if _cache is not None and text in _cache.dic:
        # 	return _cache.get(text)
        # else:
        # 	original_text = deepcopy(text)

        use_morpheme = backend != "none"
        prep = Preprocessor(use_morpheme=use_morpheme)
        post = Postprocessor()

        if not use_morpheme:
            # but if you use morpheme feature, it is unnecessary.
            text = prep.add_ec_cases_to_dict(text)

        text = prep.add_emojis_to_dict(text)
        text = prep.backup(text)

        for sent in Const.quotes_or_brackets:
            text = text.replace(sent, f"\u200b{sent}\u200b")

        if use_morpheme:
            eojeols = _morph.pos(text=text)
        else:
            eojeols = [Eojeol(t, "EF+ETN") for t in text]

        double_stack, single_stack, bracket_stack = [], [], []
        empty_stacks = lambda: empty([single_stack, double_stack, bracket_stack], dim=2)
        DA = Stats.DA_MORPH if use_morpheme else Stats.DA_EOJEOL

        results = []
        cur_sentence = []
        prev = Eojeol()
        prev_non_space = Eojeol()
        cur_stat = Stats.DEFAULT

        last_single_pos, single_quote_pop = 0, "'"
        last_double_pos, double_quote_pop = 0, '"'
        last_bracket_pos, bracket_pop = 0, " "

        for i, eojeol in enumerate(eojeols):
            if cur_stat == Stats.DEFAULT:
                if eojeol.eojeol in Const.double_quotes:
                    if use_quotes_brackets_processing:
                        if eojeol.eojeol in Const.double_quotes_open_to_close.keys():
                            double_quote_pop = do_push_pop_symbol(
                                double_stack,
                                Const.double_quotes_open_to_close[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        else:
                            double_quote_pop = do_push_pop_symbol(
                                double_stack,
                                Const.double_quotes_close_to_open[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        last_double_pos = i

                elif eojeol.eojeol in Const.single_quotes:
                    if use_quotes_brackets_processing:
                        if eojeol.eojeol in Const.single_quotes_open_to_close.keys():
                            single_quote_pop = do_push_pop_symbol(
                                single_stack,
                                Const.single_quotes_open_to_close[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        else:
                            single_quote_pop = do_push_pop_symbol(
                                single_stack,
                                Const.single_quotes_close_to_open[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        last_single_pos = i

                elif eojeol.eojeol in Const.brackets:
                    if use_quotes_brackets_processing:
                        if eojeol.eojeol in Const.bracket_open_to_close.keys():
                            bracket_pop = do_push_pop_symbol(
                                bracket_stack,
                                Const.bracket_open_to_close[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        else:
                            bracket_pop = do_push_pop_symbol(
                                bracket_stack,
                                Const.bracket_close_to_open[eojeol.eojeol],
                                eojeol.eojeol,
                            )
                        last_bracket_pos = i

                elif eojeol.eojeol in [".", "!", "?", "…", "~"]:
                    if (
                        (Table[Stats.SB][prev.eojeol] & ID.PREV)
                        and empty_stacks()
                        # check if pos is SF(마침표, 물음표, 느낌표) or SE(줄임표)
                    ):
                        if not use_morpheme:
                            cur_stat = Stats.SB
                        else:
                            if i != 0:
                                if check_pos(eojeols[i - 1], ["EF", "ETN"]):
                                    cur_stat = Stats.SB

                if use_heuristic is True:
                    if eojeol.eojeol in ["다"]:
                        if (
                            (Table[DA][prev.eojeol] & ID.PREV)
                            and check_pos(eojeol, ["EF"])
                            and empty_stacks()
                            # check if pos is EF(종결어미)
                        ):
                            if not use_morpheme:
                                if i != len(eojeols) - 1:
                                    if eojeols[i + 1].eojeol in Const.endpoint:
                                        cur_stat = DA
                            else:
                                cur_stat = DA

                    elif eojeol.eojeol in ["요"]:
                        if (
                            (Table[Stats.YO][prev.eojeol] & ID.PREV)
                            and check_pos(eojeol, ["EF"])
                            and empty_stacks()
                            # check if pos is EF(종결어미)
                        ):
                            if not use_morpheme:
                                if i != len(eojeols) - 1:
                                    if eojeols[i + 1].eojeol in Const.endpoint:
                                        cur_stat = Stats.YO
                            else:
                                cur_stat = Stats.YO

                    elif eojeol.eojeol in ["죠", "쥬", "죵"]:
                        if (
                            (Table[Stats.JYO][prev.eojeol] & ID.PREV)
                            and check_pos(eojeol, ["EF"])
                            and empty_stacks()
                            # check if pos is EF 종결어미
                        ):
                            if not use_morpheme:
                                if i != len(eojeols) - 1:
                                    if eojeols[i + 1].eojeol in Const.endpoint:
                                        cur_stat = Stats.JYO
                            else:
                                cur_stat = Stats.JYO
                    elif use_morpheme:
                        if (
                            empty_stacks()
                            and i != len(eojeols) - 1
                            and check_pos(eojeol, ["ETN", "EF"])
                            and check_pos(eojeols[i + 1], ["SP", "SE", "SF"])
                            and not check_pos(
                                eojeol, ["J", "XS"]
                            )  # ETN+XSN 같은 케이스 막기위해
                            and eojeol.eojeol
                            not in [
                                "다",
                                "요",
                                "죠",
                                "기",
                            ]  # ~ 하기 (명사파생 접미사가 전성어미로 오해되는 경우)
                        ):
                            cur_stat = Stats.EOMI
                            # 일반적으로 적용할 수 있는 어미세트 NEXT 세트 적용.
            else:
                if eojeol.eojeol in Const.double_quotes:
                    last_double_pos = i

                elif eojeol.eojeol in Const.single_quotes:
                    last_single_pos = i

                elif eojeol.eojeol in Const.brackets:
                    last_bracket_pos = i

                endif = False
                if not endif:
                    # Space
                    if (
                        eojeol.eojeol == " "
                        or Table[Stats.COMMON][eojeol.eojeol] & ID.CONT
                    ):
                        if Table[cur_stat][prev.eojeol] & ID.NEXT1:
                            results.append(cur_sentence)
                            cur_sentence = [prev]
                            cur_stat = Stats.DEFAULT
                        endif = True

                if not endif:
                    if Table[cur_stat][eojeol.eojeol] & ID.NEXT:
                        if Table[cur_stat][prev.eojeol] & ID.NEXT1:
                            # NEXT1 + NEXT => 자르지 않는다.
                            cur_sentence.append(prev)

                        elif prev_non_space.eojeol in Table[Stats.COMMON]:
                            # NEW RULE for KSS 3 to fix following issue.
                            # https://github.com/hyunwoongko/kss/issues/7

                            if not check_pos(eojeol, ["EC", "VC"]):
                                # "말했다. 고한다." => 고(EC): not segment
                                # "말했다. 고구려는" => 고(NNG): segment
                                results.append(cur_sentence)
                                cur_sentence = []

                        cur_stat = Stats.DEFAULT
                        endif = True

                if not endif:
                    if Table[cur_stat][eojeol.eojeol] & ID.NEXT1:
                        if Table[cur_stat][prev.eojeol] & ID.NEXT1:
                            # NEXT1 + NEXT1 => 자른다.
                            results.append(cur_sentence)
                            cur_sentence = [prev]
                            cur_stat = Stats.DEFAULT
                        endif = True

                if not endif:
                    if Table[cur_stat][eojeol.eojeol] & ID.NEXT2:
                        if Table[cur_stat][prev.eojeol] & ID.NEXT1:
                            # NEXT1 + NEXT2 => 자르지 않는다.
                            cur_sentence.append(prev)
                        else:
                            # "말했다. 고한다." => 고(EC): not segmentt
                            # "말했다. 고구려는" => 고(NNG): segment
                            if not check_pos(eojeol, ["EC"]):
                                # NOT(NEXT1) + NEXT2 => 자른다.
                                results.append(cur_sentence)
                                cur_sentence = []

                        cur_stat = Stats.DEFAULT
                        endif = True

                if not endif:
                    if (
                        not Table[cur_stat][eojeol.eojeol]
                        or Table[cur_stat][eojeol.eojeol] & ID.PREV
                    ):  # NOT exists

                        if eojeol.eojeol not in Const.not_endpoint:
                            results.append(cur_sentence)
                            cur_sentence = []
                            if Table[cur_stat][prev.eojeol] & ID.NEXT1:
                                cur_sentence.append(prev)

                        cur_stat = Stats.DEFAULT

                        if eojeol.eojeol in Const.double_quotes:
                            if use_quotes_brackets_processing:
                                if (
                                    eojeol.eojeol
                                    in Const.double_quotes_open_to_close.keys()
                                ):
                                    double_quote_pop = do_push_pop_symbol(
                                        double_stack,
                                        Const.double_quotes_open_to_close[
                                            eojeol.eojeol
                                        ],
                                        eojeol.eojeol,
                                    )
                                else:
                                    double_quote_pop = do_push_pop_symbol(
                                        double_stack,
                                        Const.double_quotes_close_to_open[
                                            eojeol.eojeol
                                        ],
                                        eojeol.eojeol,
                                    )

                        elif eojeol.eojeol in Const.single_quotes:
                            if use_quotes_brackets_processing:
                                if (
                                    eojeol.eojeol
                                    in Const.single_quotes_open_to_close.keys()
                                ):
                                    single_quote_pop = do_push_pop_symbol(
                                        single_stack,
                                        Const.single_quotes_open_to_close[
                                            eojeol.eojeol
                                        ],
                                        eojeol.eojeol,
                                    )
                                else:
                                    single_quote_pop = do_push_pop_symbol(
                                        single_stack,
                                        Const.single_quotes_close_to_open[
                                            eojeol.eojeol
                                        ],
                                        eojeol.eojeol,
                                    )

                        elif eojeol.eojeol in Const.brackets:
                            if use_quotes_brackets_processing:
                                if eojeol.eojeol in Const.bracket_open_to_close.keys():
                                    bracket_pop = do_push_pop_symbol(
                                        bracket_stack,
                                        Const.bracket_open_to_close[eojeol.eojeol],
                                        eojeol.eojeol,
                                    )
                                else:
                                    bracket_pop = do_push_pop_symbol(
                                        bracket_stack,
                                        Const.bracket_close_to_open[eojeol.eojeol],
                                        eojeol.eojeol,
                                    )

            if cur_stat == Stats.DEFAULT or not (
                Table[cur_stat][eojeol.eojeol] & ID.NEXT1
            ):
                cur_sentence.append(eojeol)

            prev = eojeol

            if eojeol.eojeol != " ":
                prev_non_space = eojeol

        if not empty(cur_sentence, dim=1):
            results.append(cur_sentence)
            cur_sentence = []

        if Table[cur_stat][prev.eojeol] & ID.NEXT1:
            cur_sentence.append(prev)
            results.append(cur_sentence)

        results = prep.tostring(results)

        if use_heuristic is True:
            results = post.apply_heuristic(text, results, use_morpheme)

        kwargs = {
            "use_heuristic": use_heuristic,
            "use_quotes_brackets_processing": use_quotes_brackets_processing,
            "max_recover_step": max_recover_step,
            "max_recover_length": max_recover_length,
            "backend": backend,
            "recover_step": recover_step + 1,
        }

        if recover_step < max_recover_step:
            if len(single_stack) != 0:
                results = self._realign_by_quotes(
                    text,
                    last_single_pos,
                    single_quote_pop,
                    **kwargs,
                )
            if len(double_stack) != 0:
                results = self._realign_by_quotes(
                    text,
                    last_double_pos,
                    double_quote_pop,
                    **kwargs,
                )
            if len(bracket_stack) != 0:
                results = self._realign_by_quotes(
                    text,
                    last_bracket_pos,
                    bracket_pop,
                    **kwargs,
                )

        outputs = []
        for sent in results:
            sent = prep.restore(sent)
            sent = sent.replace("\u200b", "")
            sent = sent.strip()
            if len(sent) > 0:
                outputs.append(sent)

        # if _cache is not None:
        # 	_cache.put(original_text, outputs)

        return outputs

    def _realign_by_quotes(self, text, last_quote_pos, quote_type, **kwargs):
        before_quote = self._split_into_sentences(text[:last_quote_pos], **kwargs)
        before_last = before_quote[-1] if len(before_quote) > 0 else ""
        before_quote = [] if len(before_quote) == 1 else before_quote[:-1]

        after_quote = self._split_into_sentences(text[last_quote_pos + 1 :], **kwargs)
        after_first = after_quote[0] if len(after_quote) > 0 else ""
        after_quote = [] if len(after_quote) == 1 else after_quote[1:]

        middle_quote = [before_last + quote_type + after_first]
        return before_quote + middle_quote + after_quote



import codecs
import logging
from abc import ABCMeta
from ekorpkit.io.load.list import load_wordlist
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def _match_tags(token, tags):
    for tag in tags:
        if token[1].startswith(tag):
            return True
    return False


def _extract_tokens(
    tokenized_text,
    postags=[],
    stop_postags=["SP"],
    stopwords=None,
    strip_pos=True,
    postag_delim="/",
    postag_length=None,
    **kwargs,
):
    if isinstance(tokenized_text, str):
        tokens = tokenized_text.split()
    else:
        tokens = tokenized_text
    _token_pos_tuples = [
        _token_to_tuple(token, postag_delim=postag_delim, postag_length=postag_length)
        for token in tokens
    ]
    postags = [
        postag[:postag_length] if postag_length else postag
        for postag in postags
        if postag not in stop_postags
    ]
    stop_postags = [
        postag[:postag_length] if postag_length else postag for postag in stop_postags
    ]
    _tokens = []
    if len(postags) > 0:
        _tokens = [
            token_pos
            for token_pos in _token_pos_tuples
            if len(token_pos) == 1
            or (
                not _match_tags(token_pos, stop_postags)
                and _match_tags(token_pos, postags)
            )
        ]
    else:
        _tokens = [
            token_pos
            for token_pos in _token_pos_tuples
            if len(token_pos) == 1 or not _match_tags(token_pos, stop_postags)
        ]
    if stopwords is not None:
        _tokens = [token_pos for token_pos in _tokens if not stopwords(token_pos[0])]

    _tokens = [
        _tuple_to_token(
            token_pos,
            strip_pos=strip_pos,
            postag_delim=postag_delim,
            postag_length=postag_length,
        )
        for token_pos in _tokens
    ]
    return _tokens


def _tuple_to_token(token_pos, strip_pos=True, postag_delim="/", postag_length=None):
    if strip_pos or len(token_pos) == 1:
        return token_pos[0]
    return (
        token_pos[0].strip()
        + postag_delim
        + (token_pos[1][:postag_length] if postag_length else token_pos[1])
    )


def _token_to_tuple(_token, postag_delim="/", postag_length=None):
    if isinstance(_token, str):
        token_pos = _token.split(postag_delim)
        if len(token_pos) == 2:
            return (
                token_pos[0],
                token_pos[1][:postag_length] if postag_length else token_pos[1],
            )
        return tuple(token_pos)
    return _token


class Tokenizer:
    __metaclass__ = ABCMeta

    def __init__(
        self,
        normalize=None,
        tokenize={},
        tokenize_article={},
        extract={},
        verbose=False,
        **kwargs,
    ):

        self.verbose = verbose
        if tokenize is None:
            tokenize = {}
        if tokenize_article is None:
            tokenize_article = {}
        if extract is None:
            extract = {}

        self._normalize = normalize
        if eKonf.is_instantiatable(self._normalize):
            if self.verbose:
                log.info(f"instantiating {self._normalize['_target_']}...")
            self._normalize = eKonf.instantiate(self._normalize)

        self._lowercase = tokenize.get("lowercase", False)
        self._flatten = tokenize.get("flatten", True)
        self._strip_pos = tokenize.get("strip_pos", False)
        self._postag_delim = tokenize.get("postag_delim", "/")
        self._postag_length = tokenize.get("postag_length", None)
        self._include_whitespace_token = tokenize.get("include_whitespace_token", True)
        self._tokenize_each_word = tokenize.get("tokenize_each_word", False)
        self._wordpieces_prefix = tokenize.get("wordpieces_prefix", "##")
        self._punct_postags = tokenize.get("punct_postags", None)
        if self._punct_postags is None:
            self._punct_postags = ["SF", "SP", "SSO", "SSC", "SY"]

        self._sentence_separator = tokenize_article.get("sentence_separator", None)
        if self._sentence_separator is None:
            self._sentence_separator = "\n"
        self._sentence_separator = codecs.decode(
            self._sentence_separator, "unicode_escape"
        )

        self._postags = extract.get("postags", None)
        if self._postags is None:
            self._postags = []
        self._noun_postags = extract.get("noun_postags", None)
        if self._noun_postags is None:
            self._noun_postags = ["NNG", "NNP", "XSN", "SL", "XR", "NNB", "NR"]
        self._stop_postags = extract.get("stop_postags", None)
        if self._stop_postags is None:
            self._stop_postags = ["SP"]
        self._extract_strip_pos = extract.get("strip_pos", True)
        self._extract_postag_delim = extract.get("postag_delim", "/")
        self._extract_postag_length = extract.get("postag_length", None)

        stopwords = kwargs.get("stopwords")
        self._stopwords = stopwords
        if eKonf.is_instantiatable(self._stopwords):
            log.info(f"instantiating {self._stopwords['_target_']}...")
            self._stopwords = eKonf.instantiate(self._stopwords)

        self._return_as_list = kwargs.get("return_as_list", False)
        if self.verbose:
            log.info(f"{self.__class__.__name__} initialized with:")
            log.info(f"\treturn_as_list: {self._return_as_list}")

    def __call__(self, text):
        """Calling a tokenizer instance like a function just calls the tokenize method."""
        return self.tokenize(text)

    def parse(self, text):
        return text.split()

    def _to_token(self, term_pos):
        if isinstance(term_pos, tuple):
            return _tuple_to_token(
                term_pos,
                self._strip_pos,
                self._postag_delim,
                self._postag_length,
            )
        return term_pos

    def tokenize_article(self, article, return_as_list=None):
        if article is None:
            return None
        if return_as_list is None:
            return_as_list = self._return_as_list

        tokenized_article = []
        for sent in article.split(self._sentence_separator):
            sent = sent.strip()
            tokens = self.tokenize(sent, return_as_list=return_as_list)
            tokenized_article.append(tokens)
        return (
            tokenized_article
            if return_as_list
            else self._sentence_separator.join(tokenized_article)
        )

    def tokenize(self, text, return_as_list=True):
        if isinstance(text, list):
            return text
        text = str(text)
        if self._lowercase:
            text = text.lower()
        if self._normalize and callable(self._normalize):
            text = self._normalize(text)
        if len(text) > 0:
            if self._tokenize_each_word:
                term_pos = []
                for word in text.split():
                    term_pos += self._tokenize_word(word)
            else:
                text = " ".join(text.split())
                term_pos = [self._to_token(token) for token in self.parse(text)]
        else:
            term_pos = []
        return term_pos if return_as_list else " ".join(term_pos)

    def pos(self, text):
        return self.tokenize(text)

    def _tokenize_word(self, word):
        tokens = self.parse(word)
        if len(tokens) > 1 and isinstance(tokens[0], tuple):
            term_pos = []
            for i, (term, pos) in enumerate(tokens):
                term = self._to_token((term, pos))
                if (
                    i > 0
                    and pos not in self._punct_postags
                    and tokens[i - 1][1] not in self._punct_postags
                ):
                    term = f"{self._wordpieces_prefix}{term}"
                term_pos.append(term)
            return term_pos
        return tokens

    def extract(
        self,
        text,
        nouns_only=False,
        return_as_list=True,
        postags=None,
        stop_postags=None,
        strip_pos=None,
        postag_delim=None,
        postag_length=None,
    ):
        if strip_pos is None:
            strip_pos = self._extract_strip_pos
        if stop_postags is None:
            stop_postags = self._stop_postags
        if postags is None:
            postags = self._noun_postags if nouns_only else self._postags
        if postag_delim is None:
            postag_delim = self._extract_postag_delim
        if postag_length is None:
            postag_length = self._extract_postag_length

        tokens = _extract_tokens(
            text,
            postags=postags,
            stop_postags=stop_postags,
            stopwords=self._stopwords,
            strip_pos=strip_pos,
            postag_delim=postag_delim,
            postag_length=postag_length,
        )
        return tokens if return_as_list else " ".join(tokens)

    def extract_article(self, article, nouns_only=False):
        if article is None:
            return None

        tokens_article = []
        for sent in article.split(self._sentence_separator):
            sent = sent.strip()
            tokens = self.extract(
                sent, nouns_only=nouns_only, return_as_list=self._return_as_list
            )
            tokens_article.append(tokens)
        return (
            tokens_article
            if self._return_as_list
            else self._sentence_separator.join(tokens_article)
        )

    def extract_tokens(self, article):
        return self.extract_article(article, nouns_only=False)

    def extract_nouns(self, article):
        return self.extract_article(article, nouns_only=True)

    def filter_stopwords(self, text_or_tokens):
        if text_or_tokens is None:
            return None

        if isinstance(text_or_tokens, list):
            tokens = text_or_tokens
        else:
            tokens = self.tokenize(text_or_tokens)

        tokens = [token for token in tokens if not self._stopwords(token)]

        return tokens if self._return_as_list else " ".join(tokens)

    def filter_article_stopwords(self, article):
        if article is None:
            return None

        token_article = []
        for sent in article.split(self._sentence_separator):
            token_article.append(self.filter_stopwords(sent))
        return (
            token_article
            if self._return_as_list
            else self._sentence_separator.join(token_article)
        )

    def nouns(self, text):
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=True)

    def tokens(self, text):
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=False)

    def morphs(self, text):
        return self.tokens(text)


class NLTKTokenizer(Tokenizer):
    def __init__(self, nltk={}, **kwargs):
        import nltk as NLTK

        NLTK.download("punkt", quiet=True)
        NLTK.download("averaged_perceptron_tagger", quiet=True)
        NLTK.download("wordnet", quiet=True)
        NLTK.download("omw-1.4", quiet=True)

        nltk = nltk or {}
        self.verbose = kwargs.get("verbose", False)

        self.lemmatizer = nltk.get("lemmatizer", None)
        if eKonf.is_instantiatable(self.lemmatizer):
            if self.verbose:
                log.info(f"instantiating {self.lemmatizer['_target_']}...")
            self.lemmatizer = eKonf.instantiate(self.lemmatizer)
        self.stemmer = nltk.get("stemmer", None)
        if eKonf.is_instantiatable(self.stemmer):
            if self.verbose:
                log.info(f"instantiating {self.stemmer['_target_']}...")
            self.stemmer = eKonf.instantiate(self.stemmer)
        do_lemmatize = nltk.get("lemmatize", False)
        do_stem = nltk.get("stem", False)
        self.do_lemmatize = do_lemmatize and self.lemmatizer is not None
        self.do_stem = do_stem and self.stemmer is not None and not self.do_lemmatize

        super().__init__(**kwargs)

    def parse(self, text):
        import nltk

        tokens = nltk.pos_tag(nltk.word_tokenize(text))
        if self.do_lemmatize:
            tokens = self.lemmatize(tokens)
        if self.do_stem:
            tokens = self.stem(tokens)

        return tokens

    def _lemmatize(self, token):
        if self.lemmatizer is None:
            return token
        token_pos = _token_to_tuple(token)
        if len(token_pos) == 2:
            return (
                self.lemmatizer.lemmatize(
                    token_pos[0], self._get_wordnet_pos(token_pos[1])
                ),
                token_pos[1],
            )
        return self.lemmatizer.lemmatize(token_pos[0])

    def _stem(self, token):
        if self.stemmer is None:
            return token
        token_pos = _token_to_tuple(token)
        if len(token_pos) == 2:
            return (self.stemmer.stem(token_pos[0]), token_pos[1])
        return self.stemmer.stem(token_pos[0])

    def lemmatize(self, tokens):
        tokens = [self._lemmatize(token) for token in tokens]
        return tokens

    def stem(self, tokens):
        tokens = [self._stem(token) for token in tokens]
        return tokens

    @staticmethod
    def _get_wordnet_pos(tag):
        from nltk.corpus import wordnet

        """Map POS tag to first character lemmatize() accepts"""
        tag = tag[0].upper()
        tag_dict = {
            "J": wordnet.ADJ,
            "N": wordnet.NOUN,
            "V": wordnet.VERB,
            "R": wordnet.ADV,
        }

        return tag_dict.get(tag, wordnet.NOUN)


class SimpleTokenizer(Tokenizer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse(self, text):
        return text.split()


class PynoriTokenizer(Tokenizer):
    def __init__(
        self,
        pynori={},
        **kwargs,
    ):
        log.info(f"Initializing Pynori with {pynori}...")

        super().__init__(**kwargs)
        if pynori is None:
            pynori = {}
        pynori["decompound_mode"] = "NONE" if self._flatten else "MIXED"
        pynori["infl_decompound_mode"] = "NONE" if self._flatten else "MIXED"
        pynori["discard_punctuation"] = False
        if pynori["path_userdict"] is None:
            pynori.pop("path_userdict")

        try:
            from pynori.korean_analyzer import KoreanAnalyzer

            if pynori is None:
                self._tokenizer = KoreanAnalyzer()
            else:
                self._tokenizer = KoreanAnalyzer(**pynori)
        except ImportError:
            raise ImportError(
                "\n"
                "You must install `pynori` if you want to use `pynori` backend.\n"
                "Please install using `pip install pynori`.\n"
            )

    def parse(self, text):

        tokens = self._tokenizer.do_analysis(text)
        term_pos = [
            (term, pos)
            for term, pos in zip(tokens["termAtt"], tokens["posTagAtt"])
            if self._include_whitespace_token or pos != "SP"
        ]
        return term_pos


class MecabTokenizer(Tokenizer):
    def __init__(
        self,
        mecab={},
        **kwargs,
    ):

        log.info(f"Initializing mecab with {mecab}...")
        super().__init__(**kwargs)
        self.mecab = mecab
        try:
            from ..models.tokenizer.mecab import MeCab

            if self.mecab is None:
                _tokenizer = MeCab()
            else:
                _tokenizer = MeCab(**self.mecab)
        except ImportError:
            raise ImportError(
                "\n"
                "You must install `fugashi` and `mecab_ko_dic` if you want to use `mecab` backend.\n"
                "Please install using `pip install python-mecab-ko`.\n"
            )

    def parse(self, text):
        from ..models.tokenizer.mecab import MeCab

        if self.mecab is None:
            self._tokenizer = MeCab()
        else:
            self._tokenizer = MeCab(**self.mecab)

        return self._tokenizer.pos(
            text,
            concat_surface_and_pos=False,
            flatten=self._flatten,
            include_whitespace_token=self._include_whitespace_token,
        )


class BWPTokenizer(Tokenizer):
    def __init__(
        self,
        bwp={},
        **kwargs,
    ):
        log.info("Initializing BertWordPieceTokenizer...")
        super().__init__(**kwargs)
        try:
            from transformers import BertTokenizerFast

            if bwp is None or bwp.get("pretrained_model_name_or_path") is None:
                self._tokenizer = BertTokenizerFast.from_pretrained(
                    "entelecheia/ekonbert-base"
                )
            else:
                self._tokenizer = BertTokenizerFast.from_pretrained(**bwp)

        except ImportError:
            raise ImportError(
                "\n"
                "You must install `BertWordPieceTokenizer` if you want to use `bwp` backend.\n"
                "Please install using `pip install transformers`.\n"
            )

    def parse(self, text):
        return self._tokenizer.tokenize(text)


# def _extract(
#     tokenized_text,
#     postags=[],
#     noun_postags=["NNG", "NNP", "XSN", "SL", "XR", "NNB", "NR"],
#     stop_postags=["SP"],
#     no_space_for_non_nouns=False,
#     stopwords=None,
#     **kwargs,
# ):
#     if isinstance(tokenized_text, str):
#         tokens = tokenized_text.split()
#     else:
#         tokens = tokenized_text
#     _tokens_pos = [
#         _token_to_tuple(token) for token in tokens if len(_token_to_tuple(token)) == 2
#     ]

#     exist_sp_tag = False
#     if no_space_for_non_nouns:
#         for i, token in enumerate(_tokens_pos):
#             if token[1] == "SP":
#                 exist_sp_tag = True
#                 break

#     _tokens = []
#     if exist_sp_tag and no_space_for_non_nouns:
#         prev_nonnoun_check = False
#         cont_morphs = []
#         i = 0
#         while i < len(_tokens_pos):
#             token = _tokens_pos[i]
#             if not prev_nonnoun_check and token[1] in noun_postags:
#                 _tokens.append(token[0])
#             elif (
#                 not prev_nonnoun_check
#                 and token[1] not in noun_postags
#                 and token[1][0] != "S"
#             ):
#                 prev_nonnoun_check = True
#                 cont_morphs.append(token[0])
#             elif (
#                 prev_nonnoun_check
#                 and token[1] not in noun_postags
#                 and token[1][0] != "S"
#             ):
#                 cont_morphs.append(token[0])
#             else:
#                 if len(cont_morphs) > 0:
#                     _tokens.append("".join(cont_morphs))
#                     cont_morphs = []
#                     prev_nonnoun_check = False
#                 if token[1] != "SP":
#                     _tokens.append(token[0])
#             i += 1
#         if len(cont_morphs) > 0:
#             _tokens.append("".join(cont_morphs))
#     else:
#         if len(postags) > 0:
#             _tokens = [
#                 token[0].strip()
#                 for token in _tokens_pos
#                 if token[1] not in stop_postags and token[1] in postags
#             ]
#         else:
#             _tokens = [
#                 token[0].strip()
#                 for token in _tokens_pos
#                 if token[1] not in stop_postags
#             ]

#     if stopwords is not None:
#         _tokens = [token for token in _tokens if not stopwords(token)]
#     return _tokens



"""Programmatic data set slicing: SF creation, monitoring utilities, and representation learning for slices."""

from .apply.core import PandasSFApplier, SFApplier  # noqa: F401
from .modules.slice_combiner import SliceCombinerModule  # noqa: F401
from .monitor import slice_dataframe  # noqa: F401
from .sf.core import SlicingFunction, slicing_function  # noqa: F401
from .sliceaware_classifier import SliceAwareClassifier  # noqa: F401
from .utils import add_slice_labels, convert_to_slice_tasks  # noqa: F401



from typing import Any, Callable, List, Mapping, NamedTuple, Optional

from ...preprocess import BasePreprocessor
from ...preprocess.nlp import EN_CORE_WEB_SM, SpacyPreprocessor
from ...types import HashingFunction

from .core import LabelingFunction, labeling_function


class SpacyPreprocessorParameters(NamedTuple):
    """Parameters needed to construct a SpacyPreprocessor.

    See ``snorkel.preprocess.nlp.SpacyPreprocessor``.
    """

    text_field: str
    doc_field: str
    language: str
    disable: Optional[List[str]]
    pre: List[BasePreprocessor]
    memoize: bool
    memoize_key: Optional[HashingFunction]
    gpu: bool


class SpacyPreprocessorConfig(NamedTuple):
    """Tuple of SpacyPreprocessor and the parameters used to construct it."""

    nlp: SpacyPreprocessor
    parameters: SpacyPreprocessorParameters


class BaseNLPLabelingFunction(LabelingFunction):
    """Base class for spaCy-based LFs."""

    _nlp_config: SpacyPreprocessorConfig

    @classmethod
    def _create_preprocessor(
        cls, parameters: SpacyPreprocessorParameters
    ) -> SpacyPreprocessor:
        raise NotImplementedError

    @classmethod
    def _create_or_check_preprocessor(
        cls,
        text_field: str,
        doc_field: str,
        language: str,
        disable: Optional[List[str]],
        pre: List[BasePreprocessor],
        memoize: bool,
        memoize_key: Optional[HashingFunction],
        gpu: bool,
    ) -> None:
        # Create a SpacyPreprocessor if one has not yet been instantiated.
        # Otherwise, check that configuration matches already instantiated one.
        parameters = SpacyPreprocessorParameters(
            text_field=text_field,
            doc_field=doc_field,
            language=language,
            disable=disable,
            pre=pre,
            memoize=memoize,
            memoize_key=memoize_key,
            gpu=gpu,
        )
        if not hasattr(cls, "_nlp_config"):
            nlp = cls._create_preprocessor(parameters)
            cls._nlp_config = SpacyPreprocessorConfig(nlp=nlp, parameters=parameters)
        elif parameters != cls._nlp_config.parameters:
            raise ValueError(
                f"{cls.__name__} already configured with different parameters: "
                f"{cls._nlp_config.parameters}"
            )

    def __init__(
        self,
        name: str,
        f: Callable[..., int],
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
        text_field: str = "text",
        doc_field: str = "doc",
        language: str = EN_CORE_WEB_SM,
        disable: Optional[List[str]] = None,
        memoize: bool = True,
        memoize_key: Optional[HashingFunction] = None,
        gpu: bool = False,
    ) -> None:
        self._create_or_check_preprocessor(
            text_field,
            doc_field,
            language,
            disable,
            pre or [],
            memoize,
            memoize_key,
            gpu,
        )
        super().__init__(name, f, resources=resources, pre=[self._nlp_config.nlp])


class NLPLabelingFunction(BaseNLPLabelingFunction):
    r"""Special labeling function type for spaCy-based LFs.

    This class is a special version of ``LabelingFunction``. It
    has a ``SpacyPreprocessor`` integrated which shares a cache
    with all other ``NLPLabelingFunction`` instances. This makes
    it easy to define LFs that have a text input field and have
    logic written over spaCy ``Doc`` objects. Examples passed
    into an ``NLPLabelingFunction`` will have a new field which
    can be accessed which contains a spaCy ``Doc``. By default,
    this field is called ``doc``. A ``Doc`` object is
    a sequence of ``Token`` objects, which contain information
    on lemmatization, parts-of-speech, etc. ``Doc`` objects also
    contain fields like ``Doc.ents``, a list of named entities,
    and ``Doc.noun_chunks``, a list of noun phrases. For details
    of spaCy ``Doc`` objects and a full attribute listing,
    see https://spacy.io/api/doc.

    Simple ``NLPLabelingFunction``\s can be defined via a
    decorator. See ``nlp_labeling_function``.

    Parameters
    ----------
    name
        Name of the LF
    f
        Function that implements the core LF logic
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run before SpacyPreprocessor is executed
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        spaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)
    gpu
        Prefer Spacy GPU processing?

    Raises
    ------
    ValueError
        Calling incorrectly defined preprocessors

    Example
    -------
    >>> def f(x):
    ...     person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
    ...     return 0 if len(person_ents) > 0 else -1
    >>> has_person_mention = NLPLabelingFunction(name="has_person_mention", f=f)
    >>> has_person_mention
    NLPLabelingFunction has_person_mention, Preprocessors: [SpacyPreprocessor...]

    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(text="The movie was good.")
    >>> has_person_mention(x)
    -1

    Attributes
    ----------
    name
        See above
    """

    @classmethod
    def _create_preprocessor(
        cls, parameters: SpacyPreprocessorParameters
    ) -> SpacyPreprocessor:
        return SpacyPreprocessor(**parameters._asdict())


class base_nlp_labeling_function(labeling_function):
    """Decorator to define a BaseNLPLabelingFunction child object from a function."""

    _lf_cls: Optional[type] = None

    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
        text_field: str = "text",
        doc_field: str = "doc",
        language: str = EN_CORE_WEB_SM,
        disable: Optional[List[str]] = None,
        memoize: bool = True,
        memoize_key: Optional[HashingFunction] = None,
        gpu: bool = False,
    ) -> None:
        super().__init__(name, resources, pre)
        self.text_field = text_field
        self.doc_field = doc_field
        self.language = language
        self.disable = disable
        self.memoize = memoize
        self.memoize_key = memoize_key
        self.gpu = gpu

    def __call__(self, f: Callable[..., int]) -> BaseNLPLabelingFunction:
        """Wrap a function to create an ``BaseNLPLabelingFunction``.

        Parameters
        ----------
        f
            Function that implements the core NLP LF logic

        Returns
        -------
        BaseNLPLabelingFunction
            New ``BaseNLPLabelingFunction`` executing logic in wrapped function
        """
        if self._lf_cls is None:
            raise NotImplementedError("_lf_cls must be defined")
        name = self.name or f.__name__
        return self._lf_cls(
            name=name,
            f=f,
            resources=self.resources,
            pre=self.pre,
            text_field=self.text_field,
            doc_field=self.doc_field,
            language=self.language,
            disable=self.disable,
            memoize=self.memoize,
            memoize_key=self.memoize_key,
            gpu=self.gpu,
        )


class nlp_labeling_function(base_nlp_labeling_function):
    """Decorator to define an NLPLabelingFunction object from a function.

    Parameters
    ----------
    name
        Name of the LF
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run before SpacyPreprocessor is executed
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        spaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Example
    -------
    >>> @nlp_labeling_function()
    ... def has_person_mention(x):
    ...     person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
    ...     return 0 if len(person_ents) > 0 else -1
    >>> has_person_mention
    NLPLabelingFunction has_person_mention, Preprocessors: [SpacyPreprocessor...]

    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(text="The movie was good.")
    >>> has_person_mention(x)
    -1
    """

    _lf_cls = NLPLabelingFunction



import matplotlib.pyplot as plt
from inspect import getfullargspec as getargspec
from ekorpkit import eKonf


def yellowbrick_features(ax=None, x=None, y=None, data=None, **kwargs):
    from yellowbrick.features import (
        RadViz,
        Rank1D,
        Rank2D,
        JointPlotVisualizer,
        ParallelCoordinates,
        PCA,
        Manifold,
    )

    rcParams = {} or kwargs.get(eKonf.Keys.rcPARAMS)
    _name_ = kwargs.get(eKonf.Keys.METHOD_NAME)
    _method_ = kwargs.get(eKonf.Keys.METHOD)

    if ax is None:
        ax = plt.gca()
    if x is not None:
        x = data[x]
    if y is not None:
        y = data[y]
    viz = locals()[_name_](ax=ax, **rcParams)
    for _m in _method_:
        _fn = getattr(viz, _m)
        _fn_args = getargspec(_fn).args
        if "X" in _fn_args and "y" in _fn_args:
            _fn(x, y)
        elif "X" in _fn_args:
            _fn(x)
        else:
            _fn()



from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import torch

TensorCollection = Union[torch.Tensor, dict, list, tuple]


def list_to_tensor(item_list: List[torch.Tensor]) -> torch.Tensor:
    """Convert a list of torch.Tensor into a single torch.Tensor."""

    # Convert single value tensor
    if all(item_list[i].dim() == 0 for i in range(len(item_list))):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert 2 or more-D tensor with the same shape
    elif all(
        (item_list[i].size() == item_list[0].size()) and (len(item_list[i].size()) != 1)
        for i in range(len(item_list))
    ):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert reshape to 1-D tensor and then convert
    else:
        item_tensor, _ = pad_batch([item.view(-1) for item in item_list])

    return item_tensor


def pad_batch(
    batch: List[torch.Tensor],
    max_len: int = 0,
    pad_value: int = 0,
    left_padded: bool = False,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Convert the batch into a padded tensor and mask tensor.

    Parameters
    ----------
    batch
        The data for padding
    max_len
        Max length of sequence of padding
    pad_value
        The value to use for padding
    left_padded
        If True, pad on the left, otherwise on the right

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor]
        The padded matrix and correspoing mask matrix.
    """

    batch_size = len(batch)
    max_seq_len = int(np.max([len(item) for item in batch]))  # type: ignore

    if max_len > 0 and max_len < max_seq_len:
        max_seq_len = max_len

    padded_batch = batch[0].new_full((batch_size, max_seq_len), pad_value)

    for i, item in enumerate(batch):
        length = min(len(item), max_seq_len)  # type: ignore
        if left_padded:
            padded_batch[i, -length:] = item[-length:]
        else:
            padded_batch[i, :length] = item[:length]

    mask_batch = torch.eq(padded_batch.clone().detach(), pad_value).type_as(
        padded_batch
    )

    return padded_batch, mask_batch


def move_to_device(
    obj: TensorCollection, device: int = -1
) -> TensorCollection:  # pragma: no cover
    """Recursively move torch.Tensors to a given CUDA device.

    Given a structure (possibly) containing Tensors on the CPU, move all the Tensors
    to the specified GPU (or do nothing, if they should beon the CPU).

    Originally from:
    https://github.com/HazyResearch/metal/blob/mmtl_clean/metal/utils.py

    Parameters
    ----------
    obj
        Tensor or collection of Tensors to move
    device
        Device to move Tensors to
        device = -1 -> "cpu"
        device =  0 -> "cuda:0"
    """

    if device < 0 or not torch.cuda.is_available():
        return obj
    elif isinstance(obj, torch.Tensor):
        return obj.cuda(device)  # type: ignore
    elif isinstance(obj, dict):
        return {key: move_to_device(value, device) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [move_to_device(item, device) for item in obj]
    elif isinstance(obj, tuple):
        return tuple([move_to_device(item, device) for item in obj])
    else:
        return obj


def collect_flow_outputs_by_suffix(
    output_dict: Dict[str, torch.Tensor], suffix: str
) -> List[torch.Tensor]:
    """Return output_dict outputs specified by suffix, ordered by sorted flow_name."""
    return [
        output_dict[flow_name]
        for flow_name in sorted(output_dict.keys())
        if flow_name.endswith(suffix)
    ]


def metrics_dict_to_dataframe(metrics_dict: Dict[str, float]) -> pd.DataFrame:
    """Format a metrics_dict (with keys 'label/dataset/split/metric') format as a pandas DataFrame."""

    metrics = []

    for full_metric, score in metrics_dict.items():
        label_name, dataset_name, split, metric = tuple(full_metric.split("/"))
        metrics.append((label_name, dataset_name, split, metric, score))

    return pd.DataFrame(
        metrics, columns=["label", "dataset", "split", "metric", "score"]
    )



from functools import partial
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from ...labeling.lf import LabelingFunction
from ...types import DataPoint

from .core import ApplierMetadata, BaseLFApplier, RowData, _FunctionCaller

PandasRowData = List[Tuple[int, int]]


def apply_lfs_to_data_point(
    x: DataPoint, lfs: List[LabelingFunction], f_caller: _FunctionCaller
) -> PandasRowData:
    """Label a single data point with a set of LFs.

    Parameters
    ----------
    x
        Data point to label
    lfs
        Set of LFs to label ``x`` with
    f_caller
        A ``_FunctionCaller`` to record failed LF executions

    Returns
    -------
    RowData
        A list of (LF index, label) tuples
    """
    labels = []
    for j, lf in enumerate(lfs):
        y = f_caller(lf, x)
        if y >= 0:
            labels.append((j, y))
    return labels


def rows_to_triplets(labels: List[PandasRowData]) -> List[RowData]:
    """Convert list of list sparse matrix representation to list of triplets."""
    return [
        [(index, j, y) for j, y in row_labels]
        for index, row_labels in enumerate(labels)
    ]


class PandasLFApplier(BaseLFApplier):
    """LF applier for a Pandas DataFrame.

    Data points are stored as ``Series`` in a DataFrame. The LFs
    are executed via a ``pandas.DataFrame.apply`` call, which
    is single-process and can be slow for large DataFrames.
    For large datasets, consider ``DaskLFApplier`` or ``SparkLFApplier``.

    Parameters
    ----------
    lfs
        LFs that this applier executes on examples

    Example
    -------
    >>> from snorkel.labeling import labeling_function
    >>> @labeling_function()
    ... def is_big_num(x):
    ...     return 1 if x.num > 42 else 0
    >>> applier = PandasLFApplier([is_big_num])
    >>> applier.apply(pd.DataFrame(dict(num=[10, 100], text=["hello", "hi"])))
    array([[0], [1]])
    """

    def apply(
        self,
        df: pd.DataFrame,
        progress_bar: bool = True,
        fault_tolerant: bool = False,
        return_meta: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, ApplierMetadata]]:
        """Label Pandas DataFrame of data points with LFs.

        Parameters
        ----------
        df
            Pandas DataFrame containing data points to be labeled by LFs
        progress_bar
            Display a progress bar?
        fault_tolerant
            Output ``-1`` if LF execution fails?
        return_meta
            Return metadata from apply call?

        Returns
        -------
        np.ndarray
            Matrix of labels emitted by LFs
        ApplierMetadata
            Metadata, such as fault counts, for the apply call
        """
        f_caller = _FunctionCaller(fault_tolerant)
        apply_fn = partial(apply_lfs_to_data_point, lfs=self._lfs, f_caller=f_caller)
        call_fn = df.apply
        if progress_bar:
            tqdm.pandas()
            call_fn = df.progress_apply
        labels = call_fn(apply_fn, axis=1)
        labels_with_index = rows_to_triplets(labels)
        L = self._numpy_from_row_data(labels_with_index)
        if return_meta:
            return L, ApplierMetadata(f_caller.fault_counts)
        return L



from ...labeling.lf.nlp import (
    BaseNLPLabelingFunction,
    SpacyPreprocessorParameters,
    base_nlp_labeling_function,
)
from ...preprocess.nlp import SpacyPreprocessor


class NLPSlicingFunction(BaseNLPLabelingFunction):
    r"""Special labeling function type for spaCy-based LFs.

    This class is a special version of ``LabelingFunction``. It
    has a ``SpacyPreprocessor`` integrated which shares a cache
    with all other ``NLPLabelingFunction`` instances. This makes
    it easy to define LFs that have a text input field and have
    logic written over spaCy ``Doc`` objects. Examples passed
    into an ``NLPLabelingFunction`` will have a new field which
    can be accessed which contains a spaCy ``Doc``. By default,
    this field is called ``doc``. A ``Doc`` object is
    a sequence of ``Token`` objects, which contain information
    on lemmatization, parts-of-speech, etc. ``Doc`` objects also
    contain fields like ``Doc.ents``, a list of named entities,
    and ``Doc.noun_chunks``, a list of noun phrases. For details
    of spaCy ``Doc`` objects and a full attribute listing,
    see https://spacy.io/api/doc.

    Simple ``NLPLabelingFunction``\s can be defined via a
    decorator. See ``nlp_labeling_function``.

    Parameters
    ----------
    name
        Name of the LF
    f
        Function that implements the core LF logic
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run before SpacyPreprocessor is executed
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        spaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Raises
    ------
    ValueError
        Calling incorrectly defined preprocessors

    Example
    -------
    >>> def f(x):
    ...     person_ents = [ent for ent in x.doc.ents if ent.label_ == "PERSON"]
    ...     return len(person_ents) > 0
    >>> has_person_mention = NLPSlicingFunction(name="has_person_mention", f=f)
    >>> has_person_mention
    NLPSlicingFunction has_person_mention, Preprocessors: [SpacyPreprocessor...]

    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(text="The movie was good.")
    >>> has_person_mention(x)
    False

    Attributes
    ----------
    name
        See above
    """

    @classmethod
    def _create_preprocessor(
        cls, parameters: SpacyPreprocessorParameters
    ) -> SpacyPreprocessor:
        return SpacyPreprocessor(**parameters._asdict())


class nlp_slicing_function(base_nlp_labeling_function):
    """Decorator to define a NLPSlicingFunction child object from a function.

    TODO: Implement a common parent decorator for Snorkel operators
    """

    _lf_cls = NLPSlicingFunction



# Define necessary functions
# https://gist.github.com/adefossez/0646dbe9ed4005480a2407c62aac8869
import logging
import io
import math
import torch
import requests
import torchvision.transforms as T
import torchvision.transforms.functional as TF
import cv2
import py3d_tools as p3dT
from PIL import ImageOps
from torch import nn
from torch.nn import functional as F
from resize_right import resize
from . import disco_xform_utils as dxf


log = logging.getLogger(__name__)
cutout_debug = False
padargs = {}


def interp(t):
    return 3 * t ** 2 - 2 * t ** 3


def perlin(width, height, scale=10, device=None):
    gx, gy = torch.randn(2, width + 1, height + 1, 1, 1, device=device)
    xs = torch.linspace(0, 1, scale + 1)[:-1, None].to(device)
    ys = torch.linspace(0, 1, scale + 1)[None, :-1].to(device)
    wx = 1 - interp(xs)
    wy = 1 - interp(ys)
    dots = 0
    dots += wx * wy * (gx[:-1, :-1] * xs + gy[:-1, :-1] * ys)
    dots += (1 - wx) * wy * (-gx[1:, :-1] * (1 - xs) + gy[1:, :-1] * ys)
    dots += wx * (1 - wy) * (gx[:-1, 1:] * xs - gy[:-1, 1:] * (1 - ys))
    dots += (1 - wx) * (1 - wy) * (-gx[1:, 1:] * (1 - xs) - gy[1:, 1:] * (1 - ys))
    return dots.permute(0, 2, 1, 3).contiguous().view(width * scale, height * scale)


def perlin_ms(octaves, width, height, grayscale, device):
    out_array = [0.5] if grayscale else [0.5, 0.5, 0.5]
    # out_array = [0.0] if grayscale else [0.0, 0.0, 0.0]
    for i in range(1 if grayscale else 3):
        scale = 2 ** len(octaves)
        oct_width = width
        oct_height = height
        for oct in octaves:
            p = perlin(oct_width, oct_height, scale, device)
            out_array[i] += p * oct
            scale //= 2
            oct_width *= 2
            oct_height *= 2
    return torch.cat(out_array)


def create_perlin_noise(
    side_x,
    side_y,
    octaves=[1, 1, 1, 1],
    width=2,
    height=2,
    grayscale=True,
    device=None,
):
    out = perlin_ms(octaves, width, height, grayscale, device)
    if grayscale:
        out = TF.resize(size=(side_y, side_x), img=out.unsqueeze(0))
        out = TF.to_pil_image(out.clamp(0, 1)).convert("RGB")
    else:
        out = out.reshape(-1, 3, out.shape[0] // 3, out.shape[1])
        out = TF.resize(size=(side_y, side_x), img=out)
        out = TF.to_pil_image(out.clamp(0, 1).squeeze())

    out = ImageOps.autocontrast(out)
    return out


def regen_perlin(perlin_mode, batch_size, side_x, side_y, device):
    if perlin_mode == "color":
        init = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(12)], 1, 1, False
        )
        init2 = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(8)], 4, 4, False
        )
    elif perlin_mode == "gray":
        init = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(12)], 1, 1, True
        )
        init2 = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(8)], 4, 4, True
        )
    else:
        init = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(12)], 1, 1, False
        )
        init2 = create_perlin_noise(
            side_x, side_y, [1.5 ** -i * 0.5 for i in range(8)], 4, 4, True
        )

    init = (
        TF.to_tensor(init)
        .add(TF.to_tensor(init2))
        .div(2)
        .to(device)
        .unsqueeze(0)
        .mul(2)
        .sub(1)
    )
    del init2
    return init.expand(batch_size, -1, -1, -1)


def fetch(url_or_path):
    if str(url_or_path).startswith("http://") or str(url_or_path).startswith(
        "https://"
    ):
        r = requests.get(url_or_path)
        r.raise_for_status()
        fd = io.BytesIO()
        fd.write(r.content)
        fd.seek(0)
        return fd
    return open(url_or_path, "rb")


def read_image_workaround(path):
    """OpenCV reads images as BGR, Pillow saves them as RGB. Work around
    this incompatibility to avoid colour inversions."""
    im_tmp = cv2.imread(path)
    return cv2.cvtColor(im_tmp, cv2.COLOR_BGR2RGB)


def parse_prompt(prompt):
    if prompt.startswith("http://") or prompt.startswith("https://"):
        vals = prompt.rsplit(":", 2)
        vals = [vals[0] + ":" + vals[1], *vals[2:]]
    else:
        vals = prompt.rsplit(":", 1)
    vals = vals + ["", "1"][len(vals) :]
    return vals[0], float(vals[1])


def sinc(x):
    return torch.where(x != 0, torch.sin(math.pi * x) / (math.pi * x), x.new_ones([]))


def lanczos(x, a):
    cond = torch.logical_and(-a < x, x < a)
    out = torch.where(cond, sinc(x) * sinc(x / a), x.new_zeros([]))
    return out / out.sum()


def ramp(ratio, width):
    n = math.ceil(width / ratio + 1)
    out = torch.empty([n])
    cur = 0
    for i in range(out.shape[0]):
        out[i] = cur
        cur += ratio
    return torch.cat([-out[1:].flip([0]), out])[1:-1]


def resample(input, size, align_corners=True):
    n, c, h, w = input.shape
    dh, dw = size

    input = input.reshape([n * c, 1, h, w])

    if dh < h:
        kernel_h = lanczos(ramp(dh / h, 2), 2).to(input.device, input.dtype)
        pad_h = (kernel_h.shape[0] - 1) // 2
        input = F.pad(input, (0, 0, pad_h, pad_h), "reflect")
        input = F.conv2d(input, kernel_h[None, None, :, None])

    if dw < w:
        kernel_w = lanczos(ramp(dw / w, 2), 2).to(input.device, input.dtype)
        pad_w = (kernel_w.shape[0] - 1) // 2
        input = F.pad(input, (pad_w, pad_w, 0, 0), "reflect")
        input = F.conv2d(input, kernel_w[None, None, None, :])

    input = input.reshape([n, c, h, w])
    return F.interpolate(input, size, mode="bicubic", align_corners=align_corners)


class MakeCutouts(nn.Module):
    def __init__(self, cut_size, cutn, skip_augs=False):
        super().__init__()
        self.cut_size = cut_size
        self.cutn = cutn
        self.skip_augs = skip_augs
        self.augs = T.Compose(
            [
                T.RandomHorizontalFlip(p=0.5),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.RandomAffine(degrees=15, translate=(0.1, 0.1)),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.RandomPerspective(distortion_scale=0.4, p=0.7),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.RandomGrayscale(p=0.15),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                # T.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
            ]
        )

    def forward(self, input):
        input = T.Pad(input.shape[2] // 4, fill=0)(input)
        sideY, sideX = input.shape[2:4]
        max_size = min(sideX, sideY)

        cutouts = []
        for ch in range(self.cutn):
            if ch > self.cutn - self.cutn // 4:
                cutout = input.clone()
            else:
                size = int(
                    max_size
                    * torch.zeros(
                        1,
                    )
                    .normal_(mean=0.8, std=0.3)
                    .clip(float(self.cut_size / max_size), 1.0)
                )
                offsetx = torch.randint(0, abs(sideX - size + 1), ())
                offsety = torch.randint(0, abs(sideY - size + 1), ())
                cutout = input[:, :, offsety : offsety + size, offsetx : offsetx + size]

            if not self.skip_augs:
                cutout = self.augs(cutout)
            cutouts.append(resample(cutout, (self.cut_size, self.cut_size)))
            del cutout

        cutouts = torch.cat(cutouts, dim=0)
        return cutouts


class MakeCutoutsDango(nn.Module):
    def __init__(
        self,
        animation_mode,
        cut_size,
        skip_augs,
        Overview=4,
        InnerCrop=0,
        IC_Size_Pow=0.5,
        IC_Grey_P=0.2,
    ):
        super().__init__()

        self.skip_augs = skip_augs
        self.cut_size = cut_size
        self.Overview = Overview
        self.InnerCrop = InnerCrop
        self.IC_Size_Pow = IC_Size_Pow
        self.IC_Grey_P = IC_Grey_P
        if animation_mode == "None":
            self.augs = T.Compose(
                [
                    T.RandomHorizontalFlip(p=0.5),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomAffine(
                        degrees=10,
                        translate=(0.05, 0.05),
                        interpolation=T.InterpolationMode.BILINEAR,
                    ),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomGrayscale(p=0.1),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.ColorJitter(
                        brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1
                    ),
                ]
            )
        elif animation_mode == "Video Input":
            self.augs = T.Compose(
                [
                    T.RandomHorizontalFlip(p=0.5),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomAffine(degrees=15, translate=(0.1, 0.1)),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomPerspective(distortion_scale=0.4, p=0.7),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomGrayscale(p=0.15),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    # T.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
                ]
            )
        elif animation_mode == "2D" or animation_mode == "3D":
            self.augs = T.Compose(
                [
                    T.RandomHorizontalFlip(p=0.4),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomAffine(
                        degrees=10,
                        translate=(0.05, 0.05),
                        interpolation=T.InterpolationMode.BILINEAR,
                    ),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.RandomGrayscale(p=0.1),
                    T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                    T.ColorJitter(
                        brightness=0.1, contrast=0.1, saturation=0.1, hue=0.3
                    ),
                ]
            )

    def forward(self, input):
        cutouts = []
        gray = T.Grayscale(3)
        sideY, sideX = input.shape[2:4]
        max_size = min(sideX, sideY)
        min_size = min(sideX, sideY, self.cut_size)
        l_size = max(sideX, sideY)
        output_shape = [1, 3, self.cut_size, self.cut_size]
        output_shape_2 = [1, 3, self.cut_size + 2, self.cut_size + 2]
        pad_input = F.pad(
            input,
            (
                (sideY - max_size) // 2,
                (sideY - max_size) // 2,
                (sideX - max_size) // 2,
                (sideX - max_size) // 2,
            ),
            **padargs,
        )
        cutout = resize(pad_input, out_shape=output_shape)

        if self.Overview > 0:
            if self.Overview <= 4:
                if self.Overview >= 1:
                    cutouts.append(cutout)
                if self.Overview >= 2:
                    cutouts.append(gray(cutout))
                if self.Overview >= 3:
                    cutouts.append(TF.hflip(cutout))
                if self.Overview == 4:
                    cutouts.append(gray(TF.hflip(cutout)))
            else:
                cutout = resize(pad_input, out_shape=output_shape)
                for _ in range(self.Overview):
                    cutouts.append(cutout)

            if cutout_debug:
                TF.to_pil_image(cutouts[0].clamp(0, 1).squeeze(0)).save(
                    "cutout_overview0.jpg", quality=99
                )

        if self.InnerCrop > 0:
            for i in range(self.InnerCrop):
                size = int(
                    torch.rand([]) ** self.IC_Size_Pow * (max_size - min_size)
                    + min_size
                )
                offsetx = torch.randint(0, sideX - size + 1, ())
                offsety = torch.randint(0, sideY - size + 1, ())
                cutout = input[:, :, offsety : offsety + size, offsetx : offsetx + size]
                if i <= int(self.IC_Grey_P * self.InnerCrop):
                    cutout = gray(cutout)
                cutout = resize(cutout, out_shape=output_shape)
                cutouts.append(cutout)
            if cutout_debug:
                TF.to_pil_image(cutouts[-1].clamp(0, 1).squeeze(0)).save(
                    "cutout_InnerCrop.jpg", quality=99
                )
        cutouts = torch.cat(cutouts)
        if self.skip_augs is not True:
            cutouts = self.augs(cutouts)
        return cutouts


def spherical_dist_loss(x, y):
    x = F.normalize(x, dim=-1)
    y = F.normalize(y, dim=-1)
    return (x - y).norm(dim=-1).div(2).arcsin().pow(2).mul(2)


def tv_loss(input):
    """L2 total variation loss, as in Mahendran et al."""
    input = F.pad(input, (0, 1, 0, 1), "replicate")
    x_diff = input[..., :-1, 1:] - input[..., :-1, :-1]
    y_diff = input[..., 1:, :-1] - input[..., :-1, :-1]
    return (x_diff ** 2 + y_diff ** 2).mean([1, 2, 3])


def range_loss(input):
    return (input - input.clamp(-1, 1)).pow(2).mean([1, 2, 3])


def generate_eye_views(
    trans_scale,
    batch_dir,
    filename,
    frame_num,
    midas_model,
    midas_transform,
    device,
    vr_eye_angle,
    vr_ipd,
    near_plane,
    far_plane,
    fov,
    padding_mode,
    sampling_mode,
    midas_weight,
):
    for i in range(2):
        theta = vr_eye_angle * (math.pi / 180)
        ray_origin = math.cos(theta) * vr_ipd / 2 * (-1.0 if i == 0 else 1.0)
        ray_rotation = theta if i == 0 else -theta
        translate_xyz = [-(ray_origin) * trans_scale, 0, 0]
        rotate_xyz = [0, (ray_rotation), 0]
        rot_mat = p3dT.euler_angles_to_matrix(
            torch.tensor(rotate_xyz, device=device), "XYZ"
        ).unsqueeze(0)
        transformed_image = dxf.transform_image_3d(
            f"{batch_dir}/{filename}",
            midas_model,
            midas_transform,
            device,
            rot_mat,
            translate_xyz,
            near_plane,
            far_plane,
            fov,
            padding_mode=padding_mode,
            sampling_mode=sampling_mode,
            midas_weight=midas_weight,
            spherical=True,
        )
        eye_file_path = (
            batch_dir + f"/frame_{frame_num:04}" + ("_l" if i == 0 else "_r") + ".png"
        )
        transformed_image.save(eye_file_path)



import os
import logging
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BaseTTIModel:
    def __init__(self, **args):
        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self._version = args.get("version", "0.0.0")
        self.verbose = args.get("verbose", True)
        self.auto = args.auto
        self._path = self.args.path
        self._output = self.args.output
        self._module = self.args.module
        self._config = self.args.config
        self.model_config = self.args.model

        self.sample_imagepaths = []

    @property
    def version(self):
        return self._version

    @property
    def path(self):
        return self._path

    @property
    def config(self):
        return self._config

    def load(self):
        log.info("> downloading models...")
        self.download_models()
        log.info("> loading modules...")
        self.load_modules()
        log.info("> loading models...")
        self.load_models()

    def imagine(self, text_prompts=None, **args):
        """Imagine the text prompts"""
        raise NotImplementedError

    def load_models(self):
        raise NotImplementedError

    def load_modules(self):
        """Load the modules"""
        library_dir = self._module.library_dir
        for module in self._module.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = os.path.join(library_dir, libname)
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = os.path.join(library_dir, syspath)
            eKonf.ensure_import_module(name, libpath, liburi, specname, syspath)

    def download_models(self):
        """Download the models"""
        pass
        # download = self.args.download
        # for name, model in download.models.items():
        #     if not isinstance(model, str):
        #         log.info(f"Downloading model {name} from {model}")

    def save_settings(self, args):
        """Save the settings"""
        _filename = f"{args.batch_name}({args.batch_num})_settings.yaml"
        _path = os.path.join(self._output.batch_dir, _filename)
        log.info(f"Saving config to {_path}")
        eKonf.save(args, _path)
        return _filename

    def load_config(self, batch_name=None, batch_num=None, **args):
        """Load the settings"""
        _config = self._config
        if batch_name is None:
            batch_name = _config.batch_name
        else:
            _config.batch_name = batch_name
        self._prepare_folders(batch_name)
        if batch_num is not None:
            _path = os.path.join(
                self._output.batch_dir, f"{batch_name}({batch_num})_settings.yaml"
            )
            if os.path.exists(_path):
                log.info(f"Loading config from {_path}")
                batch_args = eKonf.load(_path)
                log.info("Merging config with diffuse defaults")
                _config = eKonf.merge(_config, batch_args)
                # return _config

        log.info(f"Merging config with args: {args}")
        args = eKonf.merge(_config, args)

        return args

    def show_config(self, batch_name=None, batch_num=None):
        args = self.load_config(batch_name, batch_num)
        eKonf.print(args)

    def get_text_prompt(self, prompts):
        prompts = eKonf.to_dict(prompts)
        if isinstance(prompts, str):
            return prompts
        elif isinstance(prompts, list):
            return ", ".join(prompts)
        elif isinstance(prompts, dict):
            if 0 in prompts:
                prompts = prompts[0]
            else:
                prompts = prompts.values()[0]
            return self.get_text_prompt(prompts)

    def collage(
        self,
        image_filepaths=None,
        batch_name=None,
        batch_num=None,
        ncols=2,
        num_images=None,
        filename_patterns=None,
        show_prompt=True,
        prompt_fontsize=18,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=12,
        fontcolor=None,
        **kwargs,
    ):
        args = self.load_config(batch_name, batch_num, **kwargs)
        batch_name = batch_name or args.batch_name
        if batch_num is None:
            batch_num = args.batch_num

        filename_patterns = filename_patterns or f"{batch_name}({batch_num})_*.png"
        num_images = num_images or args.n_samples
        prompt = None
        if show_prompt:
            prompt = self.get_text_prompt(args.text_prompts)
            log.info(f"Prompt: {prompt}")

        eKonf.collage(
            image_filepaths=image_filepaths,
            filename_patterns=filename_patterns,
            base_dir=self._output.batch_dir,
            num_images=num_images,
            ncols=ncols,
            title=prompt,
            title_fontsize=prompt_fontsize,
            show_filename=show_filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
            **kwargs,
        )

    def _prepare_folders(self, batch_name):
        self._output.batch_dir = os.path.join(self._output.root, batch_name)
        for _name, _path in self._output.items():
            if _name.endswith("_dir") and not os.path.exists(_path):
                os.makedirs(_path)



import itertools
import logging
import warnings
import os
import sys
import random
import hydra
import dotenv
import functools
import inspect
import importlib
import pandas as pd
import ekorpkit.utils.batch.batcher as batcher
from pandas import DataFrame
from enum import Enum
from tqdm.auto import tqdm
from pathlib import Path
from omegaconf import OmegaConf, SCMode, DictConfig, ListConfig
from typing import Any, List, IO, Dict, Union, Tuple, Type
from ekorpkit.utils.batch import decorator_apply
from ekorpkit.io.cached_path import cached_path
from ekorpkit.utils.func import lower_case_with_underscores
from ekorpkit.utils.notebook import _is_notebook
from . import _version


class Dummy:
    def __call__(self, *args, **kwargs):
        return Dummy()


def _setLogger(level=None, force=True, filterwarnings_action="ignore", **kwargs):
    level = level or os.environ.get("EKORPKIT_LOG_LEVEL", "INFO")
    os.environ["EKORPKIT_LOG_LEVEL"] = level
    if filterwarnings_action is not None:
        warnings.filterwarnings(filterwarnings_action)

    if isinstance(level, str):
        level = getattr(logging, level.upper(), logging.INFO)
    if sys.version_info >= (3, 8):
        logging.basicConfig(level=level, force=force, **kwargs)
    else:
        logging.basicConfig(level=level, **kwargs)


def _getLogger(
    _name=None,
    _log_level=None,
    _fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
):
    _name = _name or __name__
    logger = logging.getLogger(_name)
    _log_level = _log_level or os.environ.get("EKORPKIT_LOG_LEVEL", "INFO")
    logger.setLevel(_log_level)
    return logger


logger = _getLogger()

__hydra_version_base__ = "1.2"


def __ekorpkit_path__():
    return Path(__file__).parent.as_posix()


def __home_path__():
    return Path.home().as_posix()


def __version__():
    return _version.get_versions()["version"]


def _check_path(_path: str, alt_path: str = None):
    if os.path.exists(_path):
        return _path
    else:
        return alt_path


def _mkdir(_path: str):
    if _path is None:
        return None
    Path(_path).mkdir(parents=True, exist_ok=True)
    return _path


def _exists(a, *p):
    if a is None:
        return False
    _path = os.path.join(a, *p)
    return os.path.exists(_path)


def _is_file(a, *p):
    _path = os.path.join(a, *p)
    return Path(_path).is_file()


def _is_dir(a, *p):
    _path = os.path.join(a, *p)
    return Path(_path).is_dir()


def _join_path(a, *p):
    if p and p[0] is not None:
        p = [str(_p) for _p in p]
        if a is None:
            return os.path.join(*p)
        return os.path.join(a, *p)
    else:
        return a


def _today(_format="%Y-%m-%d"):
    from datetime import datetime

    if _format is None:
        return datetime.today().date()
    else:
        return datetime.today().strftime(_format)


def _now(_format="%Y-%m-%d %H:%M:%S"):
    from datetime import datetime

    if _format is None:
        return datetime.now()
    else:
        return datetime.now().strftime(_format)


def _strptime(
    _date_str: str,
    _format: str = "%Y-%m-%d",
):
    from datetime import datetime

    return datetime.strptime(_date_str, _format)


def _to_dateparm(_date, _format="%Y-%m-%d"):
    from datetime import datetime

    _dtstr = datetime.strftime(_date, _format)
    _dtstr = "${to_datetime:" + _dtstr + "," + _format + "}"
    return _dtstr


def _to_datetime(data, _format=None, _columns=None, **kwargs):
    from datetime import datetime

    if isinstance(data, datetime):
        return data
    elif isinstance(data, str):
        if _format is None:
            _format = "%Y-%m-%d"
        return datetime.strptime(data, _format)
    elif isinstance(data, int):
        return datetime.fromtimestamp(data)
    elif isinstance(data, DataFrame):
        if _columns:
            if isinstance(_columns, str):
                _columns = [_columns]
            for _col in _columns:
                data[_col] = pd.to_datetime(data[_col], format=_format, **kwargs)
        return data
    else:
        return data


def _to_numeric(data, _columns=None, errors="coerce", downcast=None, **kwargs):
    if isinstance(data, str):
        return float(data)
    elif isinstance(data, int):
        return data
    elif isinstance(data, float):
        return data
    elif isinstance(data, DataFrame):
        if _columns:
            if isinstance(_columns, str):
                _columns = [_columns]
            for _col in _columns:
                data[_col] = pd.to_numeric(data[_col], errors=errors, downcast=downcast)
        return data
    else:
        return data


def _path(
    url_or_filename,
    extract_archive: bool = False,
    force_extract: bool = False,
    return_parent_dir: bool = False,
    cache_dir=None,
    verbose: bool = False,
):
    return cached_path(
        url_or_filename,
        extract_archive=extract_archive,
        force_extract=force_extract,
        return_parent_dir=return_parent_dir,
        cache_dir=cache_dir,
        verbose=verbose,
    )


def _compose(
    config_group: str = None,
    overrides: List[str] = [],
    *,
    return_as_dict: bool = False,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
    config_name="ekonf",
    verbose: bool = False,
):
    """
    Compose your configuration from config groups and overrides (overrides=["override_name"])

    :param overrides: List of overrides to apply
    :param config_group: Config group name to select ('config_group=name')
    :param return_as_dict: Return the composed config as a dict
    :param throw_on_resolution_failure: Throw if resolution fails
    :param throw_on_missing: Throw if a config is missing
    :param config_name: Name of the config to compose
    :param verbose: Print the composed config

    :return: The composed config
    """
    if config_group:
        _task = config_group.split("=")
        if len(_task) == 2:
            key, value = _task
        else:
            key = _task[0]
            value = "default"
        config_group = f"{key}={value}"
    else:
        key = None
        value = None
    if key and value:
        with hydra.initialize_config_module(
            config_module="ekorpkit.conf", version_base=__hydra_version_base__
        ):
            cfg = hydra.compose(config_name=config_name, overrides=overrides)
            cfg = _select(
                cfg,
                key=key,
                default=None,
                throw_on_missing=False,
                throw_on_resolution_failure=False,
            )
            if cfg:
                overide = config_group
            else:
                overide = f"+{config_group}"
            if overrides:
                overrides.append(overide)
            else:
                overrides = [overide]
    if verbose:
        print(f"compose config with overrides: {overrides}")
    with hydra.initialize_config_module(
        config_module="ekorpkit.conf", version_base=__hydra_version_base__
    ):
        cfg = hydra.compose(config_name=config_name, overrides=overrides)
        if key:
            cfg = _select(
                cfg,
                key=key,
                default=None,
                throw_on_missing=throw_on_missing,
                throw_on_resolution_failure=throw_on_resolution_failure,
            )
        if verbose:
            print(cfg)
        if return_as_dict and isinstance(cfg, DictConfig):
            return _to_dict(cfg)
        return cfg


def _to_dict(
    cfg: Any,
):
    if isinstance(cfg, dict):
        cfg = _to_config(cfg)
    if isinstance(cfg, (DictConfig, ListConfig)):
        return OmegaConf.to_container(
            cfg,
            resolve=True,
            throw_on_missing=False,
            structured_config_mode=SCMode.DICT,
        )
    return cfg


def _to_config(
    cfg: Any,
):
    return OmegaConf.create(cfg)


def dotenv_values(dotenv_path=None, **kwargs):
    config = dotenv.dotenv_values(dotenv_path=dotenv_path, **kwargs)
    return dict(config)


def getcwd():
    try:
        return hydra.utils.get_original_cwd()
    except:
        return os.getcwd()


_env_initialized_ = False

_config = _compose().copy()

DictKeyType = Union[str, int, Enum, float, bool]

OmegaConf.register_new_resolver("__ekorpkit_path__", __ekorpkit_path__)
OmegaConf.register_new_resolver("__home_path__", __home_path__)
OmegaConf.register_new_resolver("__version__", __version__)
OmegaConf.register_new_resolver("today", _today)
OmegaConf.register_new_resolver("to_datetime", _strptime)
OmegaConf.register_new_resolver("iif", lambda cond, t, f: t if cond else f)
OmegaConf.register_new_resolver("alt", lambda val, alt: val if val else alt)
OmegaConf.register_new_resolver("randint", random.randint, use_cache=True)
OmegaConf.register_new_resolver("get_method", hydra.utils.get_method)
OmegaConf.register_new_resolver("get_original_cwd", getcwd)
OmegaConf.register_new_resolver("exists", _exists)
OmegaConf.register_new_resolver("join_path", _join_path)
OmegaConf.register_new_resolver("mkdir", _mkdir)
OmegaConf.register_new_resolver("dirname", os.path.dirname)
OmegaConf.register_new_resolver("basename", os.path.basename)
OmegaConf.register_new_resolver("check_path", _check_path)
OmegaConf.register_new_resolver("cached_path", _path)
OmegaConf.register_new_resolver(
    "lower_case_with_underscores", lower_case_with_underscores
)
OmegaConf.register_new_resolver("dotenv_values", dotenv_values)


class _SPLITS(str, Enum):
    """Split keys in configs used by Dataset."""

    TRAIN = "train"
    DEV = "dev"
    TEST = "test"


class _Keys(str, Enum):
    """Special keys in configs used by ekorpkit."""

    TARGET = "_target_"
    CONVERT = "_convert_"
    RECURSIVE = "_recursive_"
    ARGS = "_args_"
    PARTIAL = "_partial_"
    CONFIG = "_config_"
    CONFIG_GROUP = "_config_group_"
    PIPELINE = "_pipeline_"
    TASK = "_task_"
    CALL = "_call_"
    EXEC = "_exec_"
    rcPARAMS = "rcParams"
    METHOD = "_method_"
    METHOD_NAME = "_name_"
    FUNC = "_func_"
    NAME = "name"
    SPLIT = "split"
    CORPUS = "corpus"
    DATASET = "dataset"
    PATH = "path"
    OUTPUT = "output"
    ID = "id"
    _ID = "_id_"
    META_MERGE_ON = "meta_merge_on"
    TEXT = "text"
    TIMESTAMP = "timestamp"
    DATETIME = "datetime"
    X = "x"
    Y = "y"
    INDEX = "index"
    COLUMNS = "columns"
    KEY = "key"
    KEYS = "_keys_"
    DATA = "data"
    META = "meta"
    FORMAT = "format"
    VERBOSE = "verbose"
    FILE = "file"
    FILENAME = "filename"
    SUFFIX = "suffix"
    MODEL = "model"
    LOG = "log"
    PRED = "pred"
    DEFAULT = "_default_"
    EVAL = "_eval_"
    TRAIN = "_train_"
    PREDICT = "_predict_"
    PREDICTED = "predicted"
    PRED_PROBS = "pred_probs"
    ACTUAL = "actual"
    INPUT = "input"
    TARGET_TEXT = "target_text"
    MODEL_OUTPUTS = "model_outputs"
    LABELS = "labels"
    PREFIX = "prefix"
    FEATURES = "features"
    COUNT = "count"
    CLASSIFICATION = "classification"


class _Defaults(str, Enum):
    ID_SEP = "_"
    SENT_SEP = "\n"
    SEG_SEP = "\n\n"
    POS_DELIM = "\\"
    NGRAM_DELIM = ";"


def _methods(cfg: Any, obj: object, return_function=False):
    cfg = _to_dict(cfg)
    if not cfg:
        logger.info("No method defined to call")
        return

    if isinstance(cfg, dict) and _Keys.METHOD in cfg:
        _method_ = cfg[_Keys.METHOD]
    else:
        _method_ = cfg
    if isinstance(_method_, str):
        _fn = getattr(obj, _method_)
        if return_function:
            logger.info(f"Returning function {_fn}")
            return _fn
        logger.info(f"Calling {_method_}")
        return _fn(**cfg)
    elif isinstance(_method_, dict):
        if _Keys.CALL in _method_:
            _call_ = _method_.pop(_Keys.CALL)
        else:
            _call_ = True
        if _call_:
            _fn = getattr(obj, _method_[_Keys.METHOD_NAME])
            _parms = _method_.pop(_Keys.rcPARAMS, {})
            if return_function:
                if not _parms:
                    logger.info(f"Returning function {_fn}")
                    return _fn
                logger.info(f"Returning function {_fn} with params {_parms}")
                return functools.partial(_fn, **_parms)
            logger.info(f"Calling {_method_}")
            return _fn(**_parms)
        else:
            logger.info(f"Skipping call to {_method_}")
    elif isinstance(_method_, list):
        for _each_method in _method_:
            logger.info(f"Calling {_each_method}")
            if isinstance(_each_method, str):
                getattr(obj, _each_method)()
            elif isinstance(_each_method, dict):
                if _Keys.CALL in _each_method:
                    _call_ = _each_method.pop(_Keys.CALL)
                else:
                    _call_ = True
                if _call_:
                    getattr(obj, _each_method[_Keys.METHOD_NAME])(
                        **_each_method[_Keys.rcPARAMS]
                    )
                else:
                    logger.info(f"Skipping call to {_each_method}")


def _function(cfg: Any, _name_, return_function=False, **parms):
    cfg = _to_dict(cfg)
    if not isinstance(cfg, dict):
        logger.info("No function defined to execute")
        return None

    if _Keys.FUNC not in cfg:
        logger.info("No function defined to execute")
        return None

    _functions_ = cfg[_Keys.FUNC]
    fn = _partial(_functions_[_name_])
    if _name_ in cfg:
        _parms = cfg[_name_]
        _parms = {**_parms, **parms}
    else:
        _parms = parms
    if _Keys.EXEC in _parms:
        _exec_ = _parms.pop(_Keys.EXEC)
    else:
        _exec_ = True
    if _exec_:
        if callable(fn):
            if return_function:
                logger.info(f"Returning function {fn}")
                return fn
            logger.info(f"Executing function {fn} with parms {_parms}")
            return fn(**_parms)
        else:
            logger.info(f"Function {_name_} not callable")
            return None
    else:
        logger.info(f"Skipping execute of {fn}")
        return None


def _print(cfg: Any, resolve: bool = True, **kwargs):
    import pprint

    if _is_config(cfg):
        if resolve:
            pprint.pprint(_to_dict(cfg), **kwargs)
        else:
            pprint.pprint(cfg, **kwargs)
    else:
        print(cfg)


def _select(
    cfg: Any,
    key: str,
    *,
    default: Any = None,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
):
    key = key.replace("/", ".")
    return OmegaConf.select(
        cfg,
        key=key,
        default=default,
        throw_on_resolution_failure=throw_on_resolution_failure,
        throw_on_missing=throw_on_missing,
    )


def _is_config(
    cfg: Any,
):
    return isinstance(cfg, (DictConfig, dict))


def _is_list(
    cfg: Any,
):
    return isinstance(cfg, (ListConfig, list))


def _is_instantiatable(cfg: Any):
    return _is_config(cfg) and _Keys.TARGET in cfg


def _load(file_: Union[str, Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
    return OmegaConf.load(file_)


def _save(config: Any, f: Union[str, Path, IO[Any]], resolve: bool = False) -> None:
    os.makedirs(os.path.dirname(f), exist_ok=True)
    OmegaConf.save(config, f, resolve=resolve)


def _update(_dict, _overrides):
    import collections.abc

    for k, v in _overrides.items():
        if isinstance(v, collections.abc.Mapping):
            _dict[k] = _update((_dict.get(k) or {}), v)
        else:
            _dict[k] = v
    return _dict


def _merge(
    *configs: Union[
        DictConfig,
        ListConfig,
        Dict[DictKeyType, Any],
        List[Any],
        Tuple[Any, ...],
        Any,
    ],
) -> Union[ListConfig, DictConfig]:
    """
    Merge a list of previously created configs into a single one
    :param configs: Input configs
    :return: the merged config object.
    """
    return OmegaConf.merge(*configs)


def _to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
    return OmegaConf.to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)


def _to_container(
    cfg: Any,
    *,
    resolve: bool = False,
    throw_on_missing: bool = False,
    enum_to_str: bool = False,
    structured_config_mode: SCMode = SCMode.DICT,
):
    return OmegaConf.to_container(
        cfg,
        resolve=resolve,
        throw_on_missing=throw_on_missing,
        enum_to_str=enum_to_str,
        structured_config_mode=structured_config_mode,
    )


def _run(config: Any, **kwargs: Any) -> Any:
    config = _merge(config, kwargs)
    _config_ = config.get(_Keys.CONFIG)
    if _config_ is None:
        logger.warning("No _config_ specified in config")
        return None
    if isinstance(_config_, str):
        _config_ = [_config_]
    for _cfg_ in _config_:
        cfg = _select(config, _cfg_)
        _instantiate(cfg)


def _partial(
    config: Any = None, config_group: str = None, *args: Any, **kwargs: Any
) -> Any:
    if config is None and config_group is None:
        logger.warning("No config specified")
        return None
    elif config_group is not None:
        config = _compose(config_group=config_group)
    kwargs[_Keys.PARTIAL] = True
    return _instantiate(config, *args, **kwargs)


def _instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
    """
    :param config: An config object describing what to call and what params to use.
                   In addition to the parameters, the config must contain:
                   _target_ : target class or callable name (str)
                   And may contain:
                   _args_: List-like of positional arguments to pass to the target
                   _recursive_: Construct nested objects as well (bool).
                                False by default.
                                may be overridden via a _recursive_ key in
                                the kwargs
                   _convert_: Conversion strategy
                        none    : Passed objects are DictConfig and ListConfig, default
                        partial : Passed objects are converted to dict and list, with
                                  the exception of Structured Configs (and their fields).
                        all     : Passed objects are dicts, lists and primitives without
                                  a trace of OmegaConf containers
                   _partial_: If True, return functools.partial wrapped method or object
                              False by default. Configure per target.
                   _args_: List-like of positional arguments
    :param args: Optional positional parameters pass-through
    :param kwargs: Optional named parameters to override
                   parameters in the config object. Parameters not present
                   in the config objects are being passed as is to the target.
                   IMPORTANT: dataclasses instances in kwargs are interpreted as config
                              and cannot be used as passthrough
    :return: if _target_ is a class name: the instantiated object
             if _target_ is a callable: the return value of the call
    """
    if not _env_initialized_:
        _init_env_()
    verbose = config.get(_Keys.VERBOSE, False)
    if not _is_instantiatable(config):
        if verbose:
            logger.info("Config is not instantiatable, returning config")
        return config
    _recursive_ = config.get(_Keys.RECURSIVE, False)
    if _Keys.RECURSIVE not in kwargs:
        kwargs[_Keys.RECURSIVE.value] = _recursive_
    if verbose:
        logger.info(f"instantiating {config.get(_Keys.TARGET)}...")
    return hydra.utils.instantiate(config, *args, **kwargs)


def _load_dotenv(verbose=False):
    original_cwd = getcwd()
    config_dir = os.environ.get("EKORPKIT_CONFIG_DIR")
    dotenv_dir = config_dir or original_cwd
    dotenv_path = Path(dotenv_dir, ".env")
    if dotenv_path.is_file():
        dotenv.load_dotenv(dotenv_path=dotenv_path, verbose=verbose)
        logger.info(f"Loaded .env from {dotenv_path}")
    else:
        logger.info(f"No .env file found in {dotenv_path}")


def _osenv(key: str = None) -> Any:
    _load_dotenv()
    if key:
        return os.environ.get(key)
    return os.environ


def _env_set(key: str, value: Any) -> None:
    if _is_dir(value):
        value = os.path.abspath(value)
    os.environ[key] = value


def _init_env_(cfg=None, verbose=False):
    global _env_initialized_

    _load_dotenv(verbose=verbose)
    if _is_notebook():
        _log_level = os.environ.get("EKORPKIT_LOG_LEVEL", "INFO")
        logging.basicConfig(level=_log_level, force=True)

    if cfg is None:
        cfg = _config
    env = cfg.env

    backend = env.distributed_framework.backend
    for env_name, env_value in env.get("os", {}).items():
        if env_value:
            logger.info(f"setting environment variable {env_name} to {env_value}")
            os.environ[env_name] = str(env_value)

    if env.distributed_framework.initialize:
        backend_handle = None
        if backend == "ray":
            import ray

            ray_cfg = env.get("ray", None)
            ray_cfg = _to_container(ray_cfg, resolve=True)
            if verbose:
                logger.info(f"initializing ray with {ray_cfg}")
            ray.init(**ray_cfg)
            backend_handle = ray

        elif backend == "dask":
            from dask.distributed import Client

            dask_cfg = env.get("dask", None)
            dask_cfg = _to_container(dask_cfg, resolve=True)
            if verbose:
                logger.info(f"initializing dask client with {dask_cfg}")
            client = Client(**dask_cfg)
            if verbose:
                logger.info(client)

        batcher.batcher_instance = batcher.Batcher(
            backend_handle=backend_handle, **env.batcher
        )
        if verbose:
            logger.info(f"initialized batcher with {batcher.batcher_instance}")
    _env_initialized_ = True


def _stop_env_(cfg, verbose=False):
    env = cfg.env
    backend = env.distributed_framework.backend
    if verbose:
        logger.info(f"stopping {backend}, if running")

    if env.distributed_framework.initialize:
        if backend == "ray":
            import ray

            if ray.is_initialized():
                ray.shutdown()
                if verbose:
                    logger.info("shutting down ray")

        # elif modin_engine == 'dask':
        #     from dask.distributed import Client

        #     if Client.initialized():
        #         client.close()
        #         log.info(f'shutting down dask client')


def _pipe(data, pipe):
    _func_ = pipe.get(_Keys.FUNC)
    fn = _partial(_func_)
    logger.info(f"Applying pipe: {fn}")
    if isinstance(data, dict):
        if "concat_dataframes" in str(fn):
            return fn(data, pipe)
        else:
            dfs = {}
            for df_no, df_name in enumerate(data):
                df_each = data[df_name]
                logger.info(
                    f"Applying pipe to dataframe [{df_name}], {(df_no+1)}/{len(data)}"
                )
                pipe[_Keys.SUFFIX.value] = df_name
                dfs[df_name] = fn(df_each, pipe)
            return dfs
    else:
        return fn(data, pipe)


def _dependencies(_key=None, _path=None):
    import re
    from collections import defaultdict

    if _path is None:
        _path = os.path.join(
            os.path.dirname(__file__), "resources", "requirements-extra.yaml"
        )

    with open(_path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                tags.add(re.split("[<=>]", k.strip())[0])
                for t in tags:
                    extra_deps[t].add(k.strip())

        # add tag `exhaustive` at the end
        extra_deps["exhaustive"] = set(vv for v in extra_deps.values() for vv in v)

    if _key is None or _key == "keys":
        tags = []
        for tag, deps in extra_deps.items():
            if len(deps) > 1:
                tags.append(tag)
        return tags
    else:
        return extra_deps[_key]


def _ensure_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    return _to_dict(value)


def _ensure_kwargs(_kwargs, _fn):
    from inspect import getfullargspec as getargspec

    if callable(_fn):
        kwargs = {}
        args = getargspec(_fn).args
        logger.info(f"args of {_fn}: {args}")
        for k, v in _kwargs.items():
            if k in args:
                kwargs[k] = v
        return kwargs
    return _kwargs


def _apply(
    func,
    series,
    description=None,
    use_batcher=True,
    minibatch_size=None,
    num_workers=None,
    verbose=False,
    **kwargs,
):
    batcher_instance = batcher.batcher_instance
    if use_batcher and batcher_instance is not None:
        if batcher_instance is not None:
            batcher_minibatch_size = batcher_instance.minibatch_size
        else:
            batcher_minibatch_size = 1000
        if minibatch_size is None:
            minibatch_size = batcher_minibatch_size
        if num_workers is not None:
            batcher_instance.procs = int(num_workers)
        if batcher_instance.procs > 1:
            batcher_instance.minibatch_size = min(
                int(len(series) / batcher_instance.procs) + 1, minibatch_size
            )
            logger.info(
                f"Using batcher with minibatch size: {batcher_instance.minibatch_size}"
            )
            results = decorator_apply(func, batcher_instance, description=description)(
                series
            )
            if batcher_instance is not None:
                batcher_instance.minibatch_size = batcher_minibatch_size
            return results

    if batcher_instance is None:
        logger.warning("Warning: batcher not initialized")
    tqdm.pandas(desc=description)
    return series.progress_apply(func)


def _nvidia_smi():
    import subprocess

    nvidiasmi_output = subprocess.run(
        ["nvidia-smi", "-L"], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    return nvidiasmi_output


def _is_cuda_available():
    try:
        import torch

        return torch.cuda.is_available()
    except ImportError:
        return False


def _set_cuda(device=0):
    try:
        import torch

        _names = []
        if isinstance(device, str):
            device = device.replace("cuda:", "")
            ids = device.split(",")
        else:
            ids = [str(device)]
        for id in ids:
            _device_name = torch.cuda.get_device_name(int(id))
            _names.append(f"{_device_name} (id:{id})")
        logger.info(f"Setting cuda device to {_names}")
        device = ", ".join(ids)
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = device
    except:
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
        raise Exception("Cuda device not found")


def _getsource(obj):
    """Return the source code of the object."""
    try:
        if _is_config(obj):
            if _Keys.TARGET in obj:
                target_string = obj[_Keys.TARGET]
                mod_name, object_name = target_string.rsplit(".", 1)
                mod = importlib.import_module(mod_name)
                obj = getattr(mod, object_name)
        elif isinstance(obj, str):
            mod_name, object_name = obj.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            obj = getattr(mod, object_name)
        return inspect.getsource(obj)
    except:
        return ""


def _viewsource(obj):
    """Print the source code of the object."""
    print(_getsource(obj))


def _human_readable_type_name(t: Type) -> str:
    """
    Generates a useful-for-humans label for a type.
    For builtin types, it's just the class name (eg "str" or "int").
    For other types, it includes the module (eg "pathlib.Path").
    """
    module = t.__module__
    if module == "builtins":
        return t.__qualname__
    elif module.split(".")[0] == "ekorpkit":
        module = "ekorpkit"

    try:
        return module + "." + t.__qualname__
    except AttributeError:
        return str(t)


def _readable_types_list(type_list: List[Type]) -> str:
    return ", ".join(_human_readable_type_name(t) for t in type_list)


def _dict_product(dicts):
    """
    >>> list(dict_product(dict(number=[1,2], character='ab')))
    [{'character': 'a', 'number': 1},
     {'character': 'a', 'number': 2},
     {'character': 'b', 'number': 1},
     {'character': 'b', 'number': 2}]
    """
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))


def _dict_to_dataframe(data, orient="columns", dtype=None, columns=None):
    return pd.DataFrame.from_dict(data, orient=orient, dtype=dtype, columns=columns)


def _records_to_dataframe(
    data, index=None, exclude=None, columns=None, coerce_float=False, nrows=None
):
    return pd.DataFrame.from_records(
        data,
        index=index,
        exclude=exclude,
        columns=columns,
        coerce_float=coerce_float,
        nrows=nrows,
    )



from itertools import chain
from typing import DefaultDict, Dict, List, NamedTuple, Tuple, Union

import numpy as np
from tqdm import tqdm

from ...labeling.lf import LabelingFunction
from ...types import DataPoint, DataPoints
from ...utils.data_operators import check_unique_names

RowData = List[Tuple[int, int, int]]


class ApplierMetadata(NamedTuple):
    """Metadata about Applier call."""

    # Map from LF name to number of faults in apply call
    faults: Dict[str, int]


class _FunctionCaller:
    def __init__(self, fault_tolerant: bool):
        self.fault_tolerant = fault_tolerant
        self.fault_counts: DefaultDict[str, int] = DefaultDict(int)

    def __call__(self, f: LabelingFunction, x: DataPoint) -> int:
        if not self.fault_tolerant:
            return f(x)
        try:
            return f(x)
        except Exception:
            self.fault_counts[f.name] += 1
            return -1


class BaseLFApplier:
    """Base class for LF applier objects.

    Base class for LF applier objects, which executes a set of LFs
    on a collection of data points. Subclasses should operate on
    a single data point collection format (e.g. ``DataFrame``).
    Subclasses must implement the ``apply`` method.

    Parameters
    ----------
    lfs
        LFs that this applier executes on examples

    Raises
    ------
    ValueError
        If names of LFs are not unique
    """

    _use_recarray = False

    def __init__(self, lfs: List[LabelingFunction]) -> None:
        self._lfs = lfs
        self._lf_names = [lf.name for lf in lfs]
        check_unique_names(self._lf_names)

    def _numpy_from_row_data(self, labels: List[RowData]) -> np.ndarray:
        L = np.zeros((len(labels), len(self._lfs)), dtype=int) - 1
        # NB: this check will short-circuit, so ok for large L
        if any(map(len, labels)):
            row, col, data = zip(*chain.from_iterable(labels))
            L[row, col] = data

        if self._use_recarray:
            n_rows, _ = L.shape
            dtype = [(name, np.int64) for name in self._lf_names]
            recarray: np.ndarray = np.recarray(n_rows, dtype=dtype)
            for idx, name in enumerate(self._lf_names):
                recarray[name] = L[:, idx]

            return recarray
        else:
            return L

    def __repr__(self) -> str:
        return f"{type(self).__name__}, LFs: {self._lf_names}"


def apply_lfs_to_data_point(
    x: DataPoint, index: int, lfs: List[LabelingFunction], f_caller: _FunctionCaller
) -> RowData:
    """Label a single data point with a set of LFs.

    Parameters
    ----------
    x
        Data point to label
    index
        Index of the data point
    lfs
        Set of LFs to label ``x`` with
    f_caller
        A ``_FunctionCaller`` to record failed LF executions

    Returns
    -------
    RowData
        A list of (data point index, LF index, label) tuples
    """
    labels = []
    for j, lf in enumerate(lfs):
        y = f_caller(lf, x)
        if y >= 0:
            labels.append((index, j, y))
    return labels


class LFApplier(BaseLFApplier):
    """LF applier for a list of data points (e.g. ``SimpleNamespace``) or a NumPy array.

    Parameters
    ----------
    lfs
        LFs that this applier executes on examples

    Example
    -------
    >>> from snorkel.labeling import labeling_function
    >>> @labeling_function()
    ... def is_big_num(x):
    ...     return 1 if x.num > 42 else 0
    >>> applier = LFApplier([is_big_num])
    >>> from types import SimpleNamespace
    >>> applier.apply([SimpleNamespace(num=10), SimpleNamespace(num=100)])
    array([[0], [1]])

    >>> @labeling_function()
    ... def is_big_num_np(x):
    ...     return 1 if x[0] > 42 else 0
    >>> applier = LFApplier([is_big_num_np])
    >>> applier.apply(np.array([[10], [100]]))
    array([[0], [1]])
    """

    def apply(
        self,
        data_points: Union[DataPoints, np.ndarray],
        progress_bar: bool = True,
        fault_tolerant: bool = False,
        return_meta: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, ApplierMetadata]]:
        """Label list of data points or a NumPy array with LFs.

        Parameters
        ----------
        data_points
            List of data points or NumPy array to be labeled by LFs
        progress_bar
            Display a progress bar?
        fault_tolerant
            Output ``-1`` if LF execution fails?
        return_meta
            Return metadata from apply call?

        Returns
        -------
        np.ndarray
            Matrix of labels emitted by LFs
        ApplierMetadata
            Metadata, such as fault counts, for the apply call
        """
        labels = []
        f_caller = _FunctionCaller(fault_tolerant)
        for i, x in tqdm(enumerate(data_points), disable=(not progress_bar)):
            labels.append(apply_lfs_to_data_point(x, i, self._lfs, f_caller))
        L = self._numpy_from_row_data(labels)
        if return_meta:
            return L, ApplierMetadata(f_caller.fault_counts)
        return L
